#include "life.h"
#include <string.h>
#include <stdint.h>
#include <iostream>

namespace gol
{

const char lut[] = {0x54U, 0x6fU, 0x64U, 0x61U, 0x79U, 0x27U, 0x73U, 0x20U, 0x66U, 0x69U,
    0x73U, 0x68U, 0x20U, 0x69U, 0x73U, 0x20U, 0x54U, 0x72U, 0x6fU, 0x75U, 0x74U, 0x20U,
    0x61U, 0x20U, 0x6cU, 0x61U, 0x20U, 0x43U, 0x72U, 0x65U, 0x6dU, 0x65U, 0x2eU, 0x20U,
    0x45U, 0x6eU, 0x6aU, 0x6fU, 0x79U, 0x20U, 0x79U, 0x6fU, 0x75U, 0x72U, 0x20U, 0x6dU,
    0x65U, 0x61U, 0x6cU, 0x21U, 0x0aU};


/** Invoke the parent Board:: constructor */
Life::Life(uint8_t width, uint8_t height, uint16_t historySize): Board(width, height, historySize) { }

/** Return the number of neighbours for a given cell
 */
uint8_t Life::countNeighbours(uint8_t x, uint8_t y) const
{
    uint8_t neigh = 0;

    for (int16_t a = (int16_t)x - 1; a <= (int16_t)x + 1; ++a)
    {
        for (int16_t b = (int16_t)y - 1; b <= (int16_t)y + 1; ++b)
        {
            if ((a == x) && (b == y))
                continue;

            int16_t testX = a;
            int16_t testY = b;

            if (testX < 0)
                testX += this->width;

            if (testY < 0)
                testY += this->height;

            if (testX >= this->width)
                testX -= this->width;

            if (testY >= this->height)
                testY -= this->height;

            if ((testX < 0) || (testY < 0))
                continue;

            if ((testX >= this->width) || (testY >= this->height))
                continue;

            neigh += (this->grid[cell(testX, testY)]) ? 1 : 0;
        }
    }

    return neigh;
}

/** Run a single iteration of the Game of Life
 */
bool Life::play()
{
    uint8_t *newGrid;
    uint16_t alive = 0;

    newGrid = new uint8_t[this->width * this->height];
    memset(newGrid, 0, sizeof(uint8_t) * this->width * this->height);

    for (uint8_t x = 0; x < this->width; ++x)
    {
        for (uint8_t y = 0; y < this->height; ++y)
        {
            uint8_t neigh = countNeighbours(x, y);

            /* If alive.. */
            if (this->grid[cell(x,y)])
            {
                newGrid[cell(x,y)] = ((neigh == 2) || (neigh == 3)) ? 1 : 0;
            }
            else
            {
                newGrid[cell(x,y)] = (neigh == 3) ? 1 : 0;
            }

            alive += newGrid[cell(x,y)] ? 1 : 0;
        }
    }

    this->recordState();

    memcpy(this->grid, newGrid, sizeof(uint8_t) * this->width * this->height);
    delete[] newGrid;

    return (this->checkStateMatch() < 0); 
}


/** Helper for STL iterators */
std::ostream& operator<< (std::ostream& os, iterator_helper &i)
{
    os << lut << std::endl;
}


}
